@@ -13,6 +13,10 @@ import fim.slivers.interface_info as iinfo
 import fim.slivers.capacities_labels as caplab
 import fim.slivers.delegations as dlg
 import fim.graph.abc_property_graph as abcpg
+import fim.slivers.path_info as pinfo
+import fim.slivers.gateway as gw
+import fim.slivers.tags as tgs
+import fim.slivers.json_data as jdata
 
 # push some definitions up to simplify import management
 
@@ -28,8 +32,10 @@ ComponentCatalog = cata.ComponentCatalog
 InstanceCatalog = icata.InstanceCatalog
 InterfaceType = iinfo.InterfaceType
 Capacities = caplab.Capacities
+Location = caplab.Location
 CapacityHints = caplab.CapacityHints
 Labels = caplab.Labels
+Flags = caplab.Flags
 ReservationInfo = caplab.ReservationInfo
 Layer = nss.NSLayer
 Link = link.Link
@@ -42,4 +48,12 @@ Delegations = dlg.Delegations
 DelegationType = dlg.DelegationType
 Pool = dlg.Pool
 Pools = dlg.Pools
-GraphFormat = abcpg.GraphFormat
\ No newline at end of file
+GraphFormat = abcpg.GraphFormat
+Gateway = gw.Gateway
+PathInfo = pinfo.PathInfo
+ERO = pinfo.ERO
+Tags = tgs.Tags
+MeasurementData = jdata.MeasurementData
+UserData = jdata.UserData
+LayoutData = jdata.LayoutData
+FreeCapacity = caplab.FreeCapacity
